
// dep.cpp
// Copyright 2022 Matthew Rickard
// This file is part of dep

// dep is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include "dep.h"
#include "logMsg.h"
#include <algorithm>
#include <sys/stat.h>
#include <iomanip>
#include "AbstractProcess.h"
#include "MacroNames.h"
#include <fstream>
#include "matches.h"
#include "Utils.h"
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include "stringify.h"
#if defined __linux__ || defined __CYGWIN__
#include <sys/wait.h>
#include <sys/resource.h>
#endif
#include <sys/sysinfo.h>
#include "clockwork.h"
#include "stream_vector.h"
#include <sys/mman.h>
#include "argv.h"

void Rule::setFolder(const std::string &s) {
  folder = s;
}

int operator < (const struct timespec &a, const struct timespec &b) {
  return a.tv_sec < b.tv_sec || a.tv_sec == b.tv_sec && a.tv_nsec < b.tv_nsec;
}

int operator > (const struct timespec &a, const struct timespec &b) {
  return a.tv_sec > b.tv_sec || a.tv_sec == b.tv_sec && a.tv_nsec > b.tv_nsec;
}

std::ostream & operator << (std::ostream &os, const struct timeval &tv) {
  return os << tv.tv_sec << '.' << std::setfill('0') << std::setw(6) << tv.tv_usec;
}

std::ostream & operator << (std::ostream &os, const struct tm &tm) {
  return os << formatDatetime("yy-mm-dd hh:mm:ss",
    tm.tm_year + 1900,
    tm.tm_mon + 1,
    tm.tm_mday,
    tm.tm_hour,
    tm.tm_min,
    tm.tm_sec,
    "", // tm.tm_zone,
    "");
}

std::ostream & operator << (std::ostream &os, const struct timespec &ts) {
#ifdef __MINGW32__
#define LOCALTIME_R_AVAILABLE 0
#else
#define LOCALTIME_R_AVAILABLE 1
#endif
#if LOCALTIME_R_AVAILABLE
  struct tm tm;
  localtime_r(&ts.tv_sec, &tm);
  os << tm;
#else
  struct tm *tm = localtime(&ts.tv_sec);
  os << *tm;
#endif
  int len = 9;
  int ns = ts.tv_nsec;
  while (ns && !(ns % 10)) {
    ns /= 10;
    len--;
  }
  if (ns)
    os << '.' << std::setfill('0') << std::setw(len) << ns;

  std::ostringstream offset;
  int ofs = tm.tm_gmtoff % 60;
  int ofm = (tm.tm_gmtoff / 60) % 60;
  int ofh = tm.tm_gmtoff / 3600;
  if (ofs >= 0)
    os << '+';
  os << std::setfill('0') << std::setw(2) << ofh;
  if (ofm || ofs) {
    os << std::setw(2) << ofm;
    if (ofs)
      os << std::setw(2) << ofs;
  }

  return os;
}

#define caseReturn(a) case a: return #a;

const char *buildStateToMacroName(int buildstate) {
  switch (buildstate) {
  caseReturn(BUILDSTATE_SOURCES_STALE);
  caseReturn(BUILDSTATE_STALE);
  caseReturn(BUILDSTATE_BUILDING);
  caseReturn(BUILDSTATE_STAT_DIFF);
  caseReturn(BUILDSTATE_REBUILT);
  caseReturn(BUILDSTATE_OK);
  caseReturn(BUILDSTATE_REBUILT_SAME);
  caseReturn(BUILDSTATE_NO_SOURCES);
  default: return "Unknown BUILDSTATE";
  }
}

int BuildFile::getFileState() {

  int minSourceState = BUILDSTATE_OK;
  if (fileflag & FILEFLAG_CMD_CHANGED) {
    //logMsg << "Using BUILDSTATE_STALE for " << filename << ' ' << logValue(fileflag) << std::endl;
    minSourceState = BUILDSTATE_STALE;
  }

  if (writeRule) {
    minSourceState = std::min(minSourceState, writeRule->getRuleState());
    if ((minSourceState == BUILDSTATE_REBUILT) && (fileflag & FILEFLAG_UNCHANGED))
      minSourceState = BUILDSTATE_REBUILT_SAME;
  }

  for (auto i = includeVector.begin(); i != includeVector.end() && minSourceState; i++)
    minSourceState = std::min(minSourceState, (*i)->getFileState());

  //logMsg << "Build state of " << filename << " is " << buildStateToMacroName(minSourceState) << std::endl;
  return minSourceState;
}

std::string pad(std::string s, int n) {
  if (s.size() < n)
    s += std::string(n - s.size(), ' ');
  return s;
}

/*

Determine a rule's (and it's targets') build state.
Calculate it from its sources build states, and if necessary, its sources and targets
stat() outputs including errors, file size and modification time.

*/

int Rule::getRuleState(int debug) {
  if (ruleflag & RULEFLAG_BUILDSTATE_VALID
   || buildstate == BUILDSTATE_REBUILT
   || buildstate == BUILDSTATE_REBUILT_SAME) {
    if (!debug)
      return buildstate;
  }
  //logMsg << "Getting a build state on " << targetVector[0]->filename << std::endl;

  ruleflag |= RULEFLAG_BUILDSTATE_VALID;

  int minSourceState = BUILDSTATE_NO_SOURCES;
  for (auto i = sourceVector.begin(); i != sourceVector.end() && minSourceState; i++) {
    int sourceState = (*i)->getFileState();
    if (debug && sourceState != BUILDSTATE_OK)
      *ruleset->outputFile->getOStream() << "     source " << (*i)->filename << " is "
        << buildStateToMacroName(sourceState) << std::endl;

    minSourceState = std::min(minSourceState, sourceState);
  }

  if (ruleset->buildStateVerbosity >= 3)
    debug = 1;

  if (ruleset->outputFile->getOStream() && debug && targetVector.size()
      && minSourceState != BUILDSTATE_SOURCES_STALE)
    *ruleset->outputFile->getOStream() << "Rule::getRuleState(): minSourceState = "
      << pad(buildStateToMacroName(minSourceState), 36) << ' '
      << targetVector.front()->filename << std::endl;

  int newbuildstate = BUILDSTATE_SOURCES_STALE;

  if (minSourceState == BUILDSTATE_SOURCES_STALE
   || minSourceState == BUILDSTATE_BUILDING
   || minSourceState == BUILDSTATE_STALE)
    newbuildstate = BUILDSTATE_SOURCES_STALE;
  else if (minSourceState == BUILDSTATE_REBUILT)
    newbuildstate = BUILDSTATE_STALE;
  else if (minSourceState == BUILDSTATE_STAT_DIFF)
    newbuildstate = BUILDSTATE_STALE;
  else if (minSourceState == BUILDSTATE_REBUILT_SAME)
    newbuildstate = BUILDSTATE_OK;

  else if (ruleset->senseNewer && minSourceState == BUILDSTATE_OK
                               || minSourceState == BUILDSTATE_NO_SOURCES) {
    if (ruleset->outputFile->getOStream() && ruleset->buildStateVerbosity >= 4)
      *ruleset->outputFile->getOStream() << "Getting a build state on "
        << targetVector[0]->filename << std::endl;
    // need to check the stat() results
    struct timespec oldestTarget = {0, 0};
    struct timespec newestSource = {1, 0};
    int oldestTargetValid = 0;
    int newestSourceValid = 0;
    int n = std::max(sourceVector.size(), targetVector.size());
    for (int i = 0; i < n; i++) {
      if (i < sourceVector.size()) {
        if (sourceVector[i]->getFileState() == BUILDSTATE_REBUILT_SAME
         || sourceVector[i]->fileflag & FILEFLAG_PSEUDOFILE) {
          // Don't check the timestamps of certain sources (rebuilds and pseudofiles)
          //lineabort(); // Note: This is not unit tested
        }
        else {
          //logMsg << "Calling calculateModTimes" << std::endl;
          sourceVector[i]->calculateModTimes();
          if (sourceVector[i]->modTimeResult) {
            if (ruleset->outputFile->getOStream() && ruleset->buildStateVerbosity >= 3)
              *ruleset->outputFile->getOStream() << "Couldn't get time of source "
                << sourceVector[i]->filename << std::endl;
          }
          else {
            if (ruleset->outputFile->getOStream() && ruleset->buildStateVerbosity >= 4)
              *ruleset->outputFile->getOStream() << "source " << sourceVector[i]->includeModTime << ' '
                << sourceVector[i]->filename << std::endl;
            if (!i || sourceVector[i]->includeModTime > newestSource) {
              if (sourceVector[i]->fileflag & FILEFLAG_EXIST) {
                //logMsg << sourceVector[i]->filename << " IS an existence dependency" << std::endl;
              }
              else {
                //logMsg << sourceVector[i]->filename << " is not an existence dependency" << std::endl;
                newestSource = sourceVector[i]->includeModTime;
              }
              newestSourceValid = 1;
              if (oldestTargetValid && newestSource > oldestTarget)
                break;
            }
          }
        }
      }
      /*else if (!newestSourceValid) {
        // If there aren't any source files, there is no newest
        // source file and there won't be a datestamp comparison. Same goes for target files.
        // UPDATE: May still need to call stat and get file existence
        break;
      }*/
      if (ruleset->outputFile->getOStream() && ruleset->buildStateVerbosity >= 4)
        *ruleset->outputFile->getOStream()
          << targetVector[0]->filename << " " << logValue(targetVector.size()) << std::endl;
      if (i < targetVector.size()) {
        if (targetVector[i]->fileflag & FILEFLAG_CMD_CHANGED) {
          if (ruleset->outputFile->getOStream() && ruleset->buildStateVerbosity >= 4)
            *ruleset->outputFile->getOStream() << targetVector[0]->filename <<
              " FILEFLAG_CMD_CHANGED; setting newbuildstate to BUILDSTATE_STALE and breaking"
              << std::endl;
          newbuildstate = BUILDSTATE_STALE;
          break;
        }
        if (targetVector[i]->fileflag & FILEFLAG_PSEUDOFILE) {
          // Don't check the timestamps of pseudofiles. They don't exist
          oldestTargetValid = 0;
          break;
        }
        else {
          if (ruleset->outputFile->getOStream() && ruleset->buildStateVerbosity >= 4)
            *ruleset->outputFile->getOStream() << "Calling calculateModTimes on "
              << targetVector[0]->filename << " is " << logValue(newbuildstate) << ' '
              << buildStateToMacroName(newbuildstate) << std::endl;
          targetVector[i]->calculateModTimes();
          if (targetVector[i]->modTimeResult) {
            if (ruleset->outputFile->getOStream() && ruleset->buildStateVerbosity >= 3)
              if (targetVector[i]->modTimeResult == EINVAL)
                *ruleset->outputFile->getOStream()
                  << targetVector[i]->filename << " failed validity check" << std::endl;
              else
                *ruleset->outputFile->getOStream() << "Couldn't get time of target "
                  << targetVector[i]->filename << std::endl;
            if (ruleset->outputFile->getOStream() && ruleset->buildStateVerbosity >= 4)
              *ruleset->outputFile->getOStream() << targetVector[0]->filename <<
                " clearing oldestTargetValid and breaking" << std::endl;
            oldestTargetValid = 0;
            break;
          }
          else {
            if (ruleset->outputFile->getOStream() && ruleset->buildStateVerbosity >= 4)
              *ruleset->outputFile->getOStream() << "target " << targetVector[i]->localModTime
                << ' ' << targetVector[i]->filename << std::endl;
            if (!i || targetVector[i]->localModTime < oldestTarget) {
              oldestTarget = targetVector[i]->localModTime;
              if (ruleset->outputFile->getOStream() && ruleset->buildStateVerbosity >= 4)
                *ruleset->outputFile->getOStream() << targetVector[0]->filename <<
                  " setting oldestTargetValid" << std::endl;
              oldestTargetValid = 1;
              if (newestSourceValid && newestSource > oldestTarget)
                break;
            }
          }
        }
      }
      else if (!oldestTargetValid) {
        //logMsg << "Breaking here" << std::endl;
        break;
      }
    }
    if (ruleset->outputFile->getOStream() && ruleset->buildStateVerbosity >= 4)
      *ruleset->outputFile->getOStream() << targetVector[0]->filename <<
        ' ' << logValue(oldestTargetValid) << std::endl;
    if (oldestTargetValid) {
      if (newestSourceValid)
        if (newestSource > oldestTarget)
          newbuildstate = BUILDSTATE_STALE;
        else {
          /*if (ruleset->outputFile->getOStream() && ruleset->buildStateVerbosity >= 3)
            *ruleset->outputFile->getOStream() << "Setting A build state of "
              << targetVector[0]->filename << " is " << logValue(newbuildstate) << ' '
              << buildStateToMacroName(newbuildstate) << std::endl;*/
          newbuildstate = BUILDSTATE_OK;
        }
      else {
        /*if (ruleset->outputFile->getOStream() && ruleset->buildStateVerbosity >= 3)
          *ruleset->outputFile->getOStream() << "Setting B build state of "
            << targetVector[0]->filename << " is " << logValue(newbuildstate) << ' '
            << buildStateToMacroName(newbuildstate) << std::endl;*/
        newbuildstate = BUILDSTATE_OK;
      }
    }
    else {
      newbuildstate = BUILDSTATE_STALE;
      if (ruleset->outputFile->getOStream() && ruleset->buildStateVerbosity >= 4)
        *ruleset->outputFile->getOStream() << "#9 Rule build state of " << targetVector[0]->filename <<
          " is " << logValue(newbuildstate) << ' ' << buildStateToMacroName(newbuildstate) << std::endl;
    }
  }
  else {
    logMsg << logValue(minSourceState) << ' ' << buildStateToMacroName(minSourceState)
      << std::endl;
    lineabort();
  }
  setState(newbuildstate);
  if (ruleset->outputFile->getOStream() && ruleset->buildStateVerbosity >= 3)
    *ruleset->outputFile->getOStream() << "Rule build state of " << targetVector[0]->filename <<
      " is " << logValue(newbuildstate) << ' ' << buildStateToMacroName(newbuildstate) << std::endl;
  return buildstate;
}

void BuildFile::calculateModTimes() {
  //logMsg << "Checkpoint " << filename << ' ' << logValue(ruleset->verbosity) << std::endl;
  if (fileflag & FILEFLAG_PSEUDOFILE) {
    return;
  }
  if (!(fileflag & FILEFLAG_LOCAL_MODTIME_VALID)) {
    struct stat statbuf;
    //logMsg << "Calling stat on " << filename << ' ' << logValue(ruleset->verbosity) << std::endl;
    int returnval = stat(filename.c_str(), &statbuf);
    if (returnval)
      modTimeResult = errno;
    else {
      modTimeResult = 0;
#ifdef __MINGW32__
      localModTime.tv_sec = statbuf.st_mtime;
      localModTime.tv_nsec = 0;
#else
      localModTime = statbuf.st_mtim;
#endif
      fileSize = statbuf.st_size;
      //logMsg << logValue(filename) << ' ' << logValue(fileSize) << std::endl;
#define ENABLE_NZ_VALIDITY_CHECK 1
#if ENABLE_NZ_VALIDITY_CHECK
      // Very simple validity check: ignore zero length files
      // Except directories, they are fine (and 0 length on Cygwin)
      if (!S_ISDIR(statbuf.st_mode)) {
        int toosmall = 0;
        if (endsWith(0, filename.c_str(), ".a")) {
          //logMsg << logValue(filename) << " ends with .a" << std::endl;
          toosmall = 8;
        }
        else {
          //logMsg << logValue(filename) << " does not end with .a" << std::endl;
        }
        if (fileSize <= toosmall && writeRule && !(fileflag & FILEFLAG_ZEROLENGTH)) {
          if (ruleset->buildStateVerbosity >= 0)
            *ruleset->outputFile->getOStream() << filename
              << " is " << fileSize << " bytes, marking as EINVAL" << std::endl;
          modTimeResult = EINVAL; // Set it to completely invalid
        }
      }
#endif
    }
    if (modTimeResult) {
      if (modTimeResult == ENOENT || modTimeResult == EINVAL) {
        if (!writeRule) {
          int foundIt = 0;
          for (std::vector<Rule *>::iterator i = readRuleVector.begin();
              i != readRuleVector.end(); i++) {
            Rule *rule = *i;
            if (rule->ruleflag & RULEFLAG_MARKED) {
              //*ruleset->outputFile->getOStream() << "(marked rule)" << std::endl;
              for (std::vector<BuildFile *>::iterator j = rule->targetVector.begin();
                  j != rule->targetVector.end(); j++) {
                BuildFile *bf2 = *j;
                *ruleset->outputFile->getOStream() << filename
                  << " (needed for " << bf2->filename
                  << ") does not exist and has no write rule" << std::endl;
                foundIt = 1;
                ruleset->failed = 1;
                break;
              }
            }
            if (foundIt)
              break;
          }
        }
      }
      else
        *ruleset->outputFile->getOStream() << filename << " error "
          << ErrnoToMacroName(modTimeResult) << std::endl;
    }
    if (ruleset->buildStateVerbosity >= 5) {
      *ruleset->outputFile->getOStream() << logValue(filename) << ' ' << logValue(modTimeResult)
        << ' ' << ErrnoToMacroName(modTimeResult) << std::endl;
    }
    fileflag |= FILEFLAG_LOCAL_MODTIME_VALID;
  }
  // There are two mod times. One is the local mod time of this file exclusively. It is used
  // to set BUILDSTATE_STAT_DIFF. The other is the include mod time. It considers all
  // included files and is used to look for sources that are newer than objects.
  if (!(fileflag & FILEFLAG_INCLUDE_MODTIME_VALID)) {
    includeModTime = localModTime;
    for (auto i = includeVector.begin(); i != includeVector.end(); i++) {
      (*i)->calculateModTimes();
      if (!((*i)->modTimeResult) && (*i)->includeModTime > includeModTime)
        includeModTime = (*i)->includeModTime;
    }
    fileflag |= FILEFLAG_INCLUDE_MODTIME_VALID;
  }

  //logMsg << filename << ' ' << logValue(modTimeResult) << std::endl;
}

int pipelines = 0;

int Rule::startx(int &resortFlag) {
  if (ruleset->optionDryrun) {
    for (auto i: commandVector)
      *ruleset->outputFile->getOStream() << i << std::endl;
    markRebuiltx();
    return 0;
  }

  for (auto target = targetVector.begin(); target != targetVector.end(); target++)
    if ((*target)->fileflag & FILEFLAG_COMPARE)
      ReadWholeFile((*target)->filename, backup[(*target)->filename]);

  if (commandVector.size()) {
    runCount = -1;
    int result = nextCommandx(resortFlag, 1);
    if (result || resortFlag)
      return result;
    buildstate = BUILDSTATE_BUILDING;
  }
  else
    markRebuiltx();
  return 0;
}

int Rule::nextCommandx(int &resortFlag, int startMore) {
  repeat:
  runCount++;
  if (runCount >= commandVector.size()) {
    int result = markRebuiltx();
    if (result)
      resortFlag = 1;
    return 0;
  }
  if (!startMore)
    return 0;
  std::string cmd = commandVector[runCount];
  if (cmd.size() && cmd[0] == '@')
    cmd = cmd.erase(0, 1);
  else {
    if (folder.size()) {
      chdir(folder.c_str());
      std::cerr << "(in " << folder << ") ";
    }
    std::cerr << cmd << std::endl;
  }
  if (cmd.size() && cmd[0] == '-') {
    ignoreExitCode = 1;
    cmd = cmd.erase(0, 1);
  }
  else
    ignoreExitCode = 0;
  pipeline.processVector.clear();

  if (!strncmp(cmd.c_str(), "export ", 7)) {
    //logMsg << "Should do an export " << logValue(cmd) << std::endl;
    cmd.erase(0, 7);
    //cmd = stripLeadingWhitespace(cmd);
    std::string::size_type equal = cmd.find('=');
    std::string name = cmd.substr(0, equal);
    std::string value = cmd.substr(equal + 1);
    name = stripWhitespace(stripLeadingWhitespace(name));
    value = stripWhitespace(stripLeadingWhitespace(value));
    /*int setenvresult =*/ setenv(name.c_str(), value.c_str(), 1);
    //logMsg << "Exported " << logValue(name) << ' ' << logValue(value)
      //<< ' ' << logValue(setenvresult) << std::endl;

    /* This is a little lazy. All rules are sharing dep's process environment,
       whereas for parallel execution, rules should have separate environments.
       So if there are rules that go:

       {
         export redback=ceiling
         command_a1
         command_a2
       }

       {
         export redback=toiletseat
         command_b1
         command_b2
       }

       That run in parallel, command_a1 and command_b1 will probably be right, but
       command_a2 will probably have redback=toiletseat instead of redback=ceiling

    */

    goto repeat;
  }
  //else
    //logMsg << "Starting a pipeline " << logValue(cmd) << std::endl;

  int result = System(cmd, ruleset->outputFile->getFd(O_RDWR | O_CREAT | O_TRUNC),
    -1, 0, SYSTEMFLAG_ASYNC, &pipeline);

  if (!result)
    pipelines++;
  if (folder.size())
#if FCHDIR_AVAILABLE
    fchdir(getInitialDir());
#else
    chdir(getInitialDir());
#endif
  return result;
}

int Rule::markRebuiltx() {
  int resortFlag = 0;
  setState(BUILDSTATE_REBUILT);
  for (auto target = targetVector.begin(); target != targetVector.end(); target++) {
    (*target)->fileflag &= ~FILEFLAG_CMD_CHANGED;
    if ((*target)->fileflag & FILEFLAG_COMPARE) {
      std::string wholeFile;
      ReadWholeFile((*target)->filename, wholeFile);
      if (backup[(*target)->filename] == wholeFile) {
        *ruleset->outputFile->getOStream() << (*target)->filename << " did not change" << std::endl;
        (*target)->fileflag |= FILEFLAG_UNCHANGED;
      }
      else {
        *ruleset->outputFile->getOStream() << (*target)->filename << " changed" << std::endl;
      }
    }
    if ((*target)->includesData) {
      //*ruleset->outputFile->getOStream() << "OMGWTFBBQ!!!" << std::endl;
      std::fstream is;
      //logMsg << logValue((*target)->fileflag) << std::endl;
      (*target)->fileflag &= ~FILEFLAG_FOUND_INCLUDES;
      //logMsg << logValue((*target)->fileflag) << std::endl;

      std::string foundFile(ruleset->findFile((*target)->filename,
        &(*target)->folders, is, "", folder.c_str()));

      (*target)->treeAddIncludes((*target)->includesData->filename.c_str(),
        &(*target)->folders, is, (*target)->includesData->workingDir.c_str());
      (*target)->markFile();
      resortFlag = 1;
    }
  }
  backup.clear();
  return resortFlag;
}

void Rule::setState(int newbuildstate) {
  ruleflag |= RULEFLAG_BUILDSTATE_VALID;
  if (newbuildstate != buildstate) {
    buildstate = newbuildstate;
    if (ruleset->outputFile->getOStream() && ruleset->buildStateVerbosity >= 2 && targetVector.size())
      *ruleset->outputFile->getOStream() << pad(buildStateToMacroName(buildstate), 36) << ' '
        << targetVector.front()->filename << std::endl;
    for (auto i = targetVector.begin(); i != targetVector.end(); i++)
      (*i)->invalidateState();
  }
}

void BuildFile::invalidateState() {
  for (auto j = readRuleVector.begin(); j != readRuleVector.end(); j++)
    (*j)->invalidateState();
  for (auto j = includedByVector.begin(); j != includedByVector.end(); j++)
    (*j)->invalidateState();
}

void Rule::invalidateState() {
  if (ruleflag & RULEFLAG_BUILDSTATE_VALID) {
    ruleflag &= ~RULEFLAG_BUILDSTATE_VALID;
    for (auto i = targetVector.begin(); i != targetVector.end(); i++)
      (*i)->invalidateState();
  }
}

void Rule::addCommand(std::string command) {
  size_t pos = command.find("%in ");
  if (pos != std::string::npos)
    command.resize(pos);
  pos = command.find("%out ");
  if (pos != std::string::npos)
    command.resize(pos);
  pos = command.find("%cmpout ");
  if (pos != std::string::npos)
    command.resize(pos);
  pos = command.find("%pseudo ");
  if (pos != std::string::npos)
    command.resize(pos);
  pos = command.find("%exist ");
  if (pos != std::string::npos)
    command.resize(pos);
  pos = command.find("%outz ");
  if (pos != std::string::npos)
    command.resize(pos);
  //logMsg << "adding command {" << logValue(command) << "} to rule " << logValue(this) << std::endl;
  if (command.size())
    commandVector.push_back(stripLeadingWhitespace(stripWhitespace(command)));
}

std::string normalizePath(std::string a) {
  std::string old = a;
  for (;;) {
    size_t doubleSlash = a.find("//");
    if (doubleSlash == std::string::npos)
      break;
    a.erase(doubleSlash, 1);
  }
  while (a.size() >= 2 && a[0] == '.' && a[1] == '/')
    a.erase(0, 2);
  for (;;) {
    size_t dot = a.find("/./");
    if (dot == std::string::npos)
      break;
    a.erase(dot, 2);
  }
  for (;;) {
    size_t doubleDot = a.find("/../");
    if (doubleDot == std::string::npos)
      break;
    size_t previousSlash = a.rfind('/', doubleDot - 1);
    if (previousSlash == -1) {
      a.erase(0, doubleDot + 4);
    }
    else {
      a.erase(previousSlash, doubleDot + 3 - previousSlash);
    }
  }
  //if (a != old)
    //logMsg << "Changed from " << old << " to " << a << std::endl;
  return a;
}

std::string joinFolderAndFile(const std::string &folder, const std::string &file) {
  if (!folder.size())
    return file;
  if (file.size() && file[0] == '/')
    return file;
  std::string result = folder + "/" + file;
  result = normalizePath(result);
  return result;
}

RuleSet::RuleSet() {
  optionTreeIncludes = 1;
  optionRuleIncludes = 0;
  optionDryrun = 0;
  optionRebuildMarked = 0;
  optionRebuildAll = 0;

  senseNewer = 1;
  senseModTime = 1;
  senseSize = 1;
  senseRule = 1;
  buildStateVerbosity = 0;
  otherVerbosity = 0;

  maxPipelines = -13;
  maxNonIncludes = -13;
  outputFile = 0;
  failed = 0;
  confirmRuleChanges = 0;
}

static DepSharedMem *sharedmem;

static sighandler_t previousHandler[_NSIG];

static std::string shmname;

void depExitHandler() {
  //logMsg << "Unlinking " << shmname << std::endl;
  shm_unlink(shmname.c_str());
}

void depSignalHandler(int sig) {
  static bool triedThis(0);
  if (triedThis) {
    logMsg << olib::argv[0] << " caught signal " << SignalToMacroName(sig)
      << ". Must have totally mangled the signal handlers. _exit()ing this time" << std::endl;
    _exit(0);
  }
  triedThis = 1;
  safeSignal(sig, previousHandler[sig]);
  logMsg << "dep caught signal " << SignalToMacroName(sig)
    << ". restored the old signal handler (" << previousHandler[sig]
    << ") and unlinking shared memory" << std::endl;

  shm_unlink(shmname.c_str());

  logMsg << "dep unlinked shared memory and re-raising the signal. This should exit the process"
    << std::endl;
  raise(sig); // this shouldn't return
  logMsg << olib::argv[0] << " Hmm. raise() returned without exiting the process. Must have totally "
    "mangled the signal handlers. _exit()ing this time" << std::endl;
  _exit(0);
}

int RuleSet::processCmd(int argc, const char *argv[], File *outputFile,
    const RuleSet *cacheset) {
  this->outputFile = outputFile;
  int foundATarget = 0;
  for (int i = 0; i < argc; i++) {
    std::map<std::string, BuildFile *>::iterator file = fileMap.find(argv[i]);
    if (file == fileMap.end()) {
      std::cerr << "dep: can't find any rule for " << argv[i] << std::endl;
      return -1;
    }
    if (file->second->writeRule) {
      file->second->writeRule->markRule();
      if (optionRebuildMarked)
        file->second->writeRule->ruleflag |= RULEFLAG_REBUILD;
    }
    else {
      std::cerr << "dep: can't find a build rule for " << argv[i] << std::endl;
      return -1;
    }
    foundATarget = 1;
  }

  if (!foundATarget) {
    if (ruleVector.size()) {
      if (ruleVector.back()->targetVector.size()) {
        std::cerr << "Targetting " << ruleVector.back()->targetVector[0]->filename
          << std::endl;
        foundATarget = 1;
      }
      else {
        std::cerr << "Targetting something but " << logValue(ruleVector.back()->targetVector.size())
          << std::endl;
        foundATarget = 1;
      }
      ruleVector.back()->markRule();
      if (optionRebuildMarked)
        ruleVector.back()->ruleflag |= RULEFLAG_REBUILD;
    }
  }
  else {
    //logMsg << logValue(ruleVector.size()) << std::endl;
  }
  if (cacheset)
    compareToCache(*cacheset);

  // Warning about files with no write rule that do not exist
  for (std::map<std::string, BuildFile *>::iterator i = fileMap.begin(); i != fileMap.end(); i++)
    if (!i->second->writeRule)
      i->second->calculateModTimes();

  if (!foundATarget) {
    std::cerr << "dep: No target selected" << std::endl;
  }

  shmname = "dep_" + stringify(getpid());

  /*int setenvresult =*/ setenv("DEP_SHM", shmname.c_str(), 1);

#ifdef unix
  previousHandler[SIGHUP] = safeSignal(SIGHUP, depSignalHandler);
  previousHandler[SIGQUIT] = safeSignal(SIGQUIT, depSignalHandler);
  previousHandler[SIGPIPE] = safeSignal(SIGPIPE, depSignalHandler);
#endif
  previousHandler[SIGINT] = safeSignal(SIGINT, depSignalHandler);
  previousHandler[SIGTERM] = safeSignal(SIGTERM, depSignalHandler);

  atexit(depExitHandler);

  int shmfd = shm_open(shmname.c_str(), O_CREAT | O_RDWR | O_TRUNC, 0644);
  /*int tresult =*/ ftruncate(shmfd, 4096);
  sharedmem = (DepSharedMem *)mmap(0, 4096, PROT_READ | PROT_WRITE, MAP_SHARED, shmfd, 0);
  //logMsg << logValue(setenvresult) << ' ' << logValue(shmfd) << ' ' << logValue(tresult)
    //<< ' ' << logValue(sharedmem) << std::endl;

  return run(outputFile);
}

static void addSourcesTo(Rule *rule, std::vector<Rule *> &v, std::vector<Rule *> &from) {
  for (auto buildFile = rule->sourceVector.begin();
      buildFile != rule->sourceVector.end(); buildFile++) {
    if ((*buildFile)->writeRule) {
      auto j = v.begin();
      for (; j != v.end(); j++)
        if (*j == (*buildFile)->writeRule) {
          //logMsg << "Breaking because " << *j << ' ' << (*j)->targetVector[0]->filename
            //<< " is the same as " << (*i)->writeRule << ' ' << (*i)->filename << std::endl;
          break;
        }
      if (j == v.end()) {
        for (auto k = from.begin(); k != from.end(); k++) {
          if ((*buildFile)->writeRule == *k) {
            //logMsg << "Adding " << (*i)->writeRule << ' ' << (*i)->filename
            //<< " as a source of " << rule << ' ' << rule->targetVector[0]->filename
            //<< std::endl;
            v.push_back((*buildFile)->writeRule);
            addSourcesTo((*buildFile)->writeRule, v, from);
            break;
          }
        }
      }
    }
#define CHECK_INCLUDEFILES 1
#if CHECK_INCLUDEFILES
    // Todo: Can't see the effect of this code in unit tests. Make sure it is unit tested!
    for (auto includeFile = (*buildFile)->includeVector.begin();
        includeFile != (*buildFile)->includeVector.end(); includeFile++) {
      // This might do one level but it should be recursive
      if ((*includeFile)->writeRule) {
        auto j = v.begin();
        for (; j != v.end(); j++)
          if (*j == (*includeFile)->writeRule) {
            //logMsg << "Breaking because " << *j << ' ' << (*j)->targetVector[0]->filename
              //<< " is the same as " << (*i)->writeRule << ' ' << (*i)->filename << std::endl;
            break;
          }
        if (j == v.end()) {
          for (auto k = from.begin(); k != from.end(); k++) {
            if ((*includeFile)->writeRule == *k) {
              //logMsg << "Adding " << (*i)->writeRule << ' ' << (*i)->filename
              //<< " as a source of " << rule << ' ' << rule->targetVector[0]->filename
              //<< std::endl;
              v.push_back((*includeFile)->writeRule);
              addSourcesTo((*includeFile)->writeRule, v, from);
              break;
            }
          }
        }
      }
    }
#endif
  }
}

void sortRuleVector(std::vector<Rule *> &ruleVector) {
  if (ruleVector.size() < 2) {
    //logMsg << "sortRuleVector: too darn trivial" << std::endl;
    return;
  }

  // That algorithm:
  // 1 choose a rule
  // 2 find all its sources
  // 3 recursively sort all of its sources
  // 4 find all its non-sources
  // 5 output itself
  // 6 recursively sort all its non-sources

  Rule *mostImportant = *ruleVector.begin();

  // 2 find all its sources
  std::vector<Rule *> sources;
  //logMsg << "Calling addSourcesTo" << std::endl;
  addSourcesTo(mostImportant, sources, ruleVector);

  // 3 recursively sort all of its sources
  sortRuleVector(sources);

  // 4 find all its non-sources
  std::vector<Rule *> nonsources;
  for (auto i = ruleVector.begin(); i != ruleVector.end(); i++) {
    if (*i == mostImportant)
      continue;
    auto j = sources.begin();;
    for (; j != sources.end(); j++) {
      if (*j == *i)
        break;
    }
    if (j == sources.end())
      nonsources.push_back(*i);
  }

  if (sources.size() + nonsources.size() + 1 != ruleVector.size()) {
    logMsg << "Oh no! sorted " << ruleVector.size() << " rules into "
      << sources.size() << " + 1 + " << nonsources.size() << std::endl;
    for (auto i = ruleVector.begin(); i != ruleVector.end(); i++)
      logMsg << (*i)->targetVector[0]->filename << std::endl;
    logMsg << "Sources:" << std::endl;
    for (auto i = sources.begin(); i != sources.end(); i++)
      logMsg << (*i)->targetVector[0]->filename << std::endl;
    logMsg << "Most important:" << std::endl;
    logMsg << mostImportant->targetVector[0]->filename << std::endl;
    logMsg << "Non-Sources:" << std::endl;
    for (auto i = nonsources.begin(); i != nonsources.end(); i++)
      logMsg << (*i)->targetVector[0]->filename << std::endl;
    logMsg << std::endl;
    lineabort();
  }

  // 5 output itself
  sources.push_back(mostImportant);

  // 6 recursively sort all its non-sources
  sortRuleVector(nonsources);

  // 7 concatentate and return
  sources.insert(sources.end(), nonsources.begin(), nonsources.end());
  std::swap(sources, ruleVector);
}

timeval &operator -= (timeval &a, const timeval &b) {
  if (a.tv_usec < b.tv_usec) {
    a.tv_usec += 1000000;
    a.tv_sec -= 1;
  }
  a.tv_sec -= b.tv_sec;
  a.tv_usec -= b.tv_usec;
  return a;
}

int RuleSet::waitForRuleExitx(int &resortFlag, int startMore) {
  int wstatus;

  reallyRestart:

#ifdef __MINGW32__
  int pid = 0;
  lineabort();
#else
  //logMsg << "Waiting" << std::endl;
  int pid = wait(&wstatus);
  //logMsg << "Wait complete" << std::endl;
#endif

  if (pid == -1) {
    int err = errno;
    if (err == ECHILD)
      return 0;
    logMsg << "errno = " << err << ' ' << ErrnoToMacroName(err) << std::endl;
    lineabort();
  }

  int foundIt = 0;
  restart:

  // Look through all processes in the current pipeline in all rules
  for (std::vector<Rule *>::iterator i = ruleVector.begin(); i != ruleVector.end(); i++) {
    //logMsg << "Looking at a rule, " << logValue((*i)->targetVector.size()) << std::endl;
    for (std::vector<Process *>::iterator j = (*i)->pipeline.processVector.begin();
        j != (*i)->pipeline.processVector.end(); j++) {
      //logMsg << logValue(*j) << std::endl;
      if (*j && pid == (*j)->child) {
        //logMsg << "Found it " << logValue(wstatus) << std::endl;
        foundIt = 1;
        if (wstatus) {
          failedCommands << (*j)->args << ": " << waitString(wstatus) << '\n';
          //logMsg << logValue(wstatus) << ' ' << logValue((*j)->args) << std::endl;
        }
        (*j)->child = -1;
        //logMsg << logValue((*i)->ignoreExitCode) << std::endl;
        if ((*i)->ignoreExitCode)
          wstatus = 0;
        // Check if any processes in the pipeline are still running
        for (std::vector<Process *>::iterator k = (*i)->pipeline.processVector.begin();
            k != (*i)->pipeline.processVector.end(); k++) {
          if ((*k)->child != -1) {
            return wstatus;
          }
        }
        // All processes in the pipeline have exited; proceed to the next pipeline
        pipelines--;
        if (!failed) {
          (*i)->nextCommandx(resortFlag, startMore);
          if (resortFlag)
            return wstatus;
          goto restart;
        }
      }
      //else {
        //logMsg << "Nup" << std::endl;
        //if (otherVerbosity >= 3)
          //*outputFile->getOStream() << "[unrecognized child process?]" << std::endl;
      //}
    }
  }
  if (!foundIt && otherVerbosity >= 3) {
    *outputFile->getOStream() << "[unrecognized child process]" << std::endl;
    logMsg << "Unrecognized child process" << std::endl;
    goto reallyRestart;
  }
  //return 0;
  return wstatus;
}

static std::string quoteAsNec(String s) {
  if (s.s[0] == '\'' || s.s[0] == '\"')
    return s.s;
  if (strchr(s.s, ' '))
    return std::string("\"") + s.s + "\"";
  return s.s;
}

int RuleSet::run(File *outputFile) {
  pipelines = 0;
  if (failed)
    return -1;
    
  this->outputFile = outputFile;
  getInitialDir();
  //this->outputFile->getOStream() = os;
  if (!outputFile)
    otherVerbosity = 0;

  resortLabel:

  // Copy out and sort the marked rules
  std::vector<Rule *> newRuleVector;
  for (std::vector<Rule *>::iterator i = ruleVector.begin(); i != ruleVector.end(); i++)
    if ((*i)->ruleflag & RULEFLAG_MARKED)
      newRuleVector.push_back(*i);
  sortRuleVector(newRuleVector);
  //*outputFile->getOStream() << logValue(verbosity) << std::endl;
  if (otherVerbosity >= 2) {
    *outputFile->getOStream() << "Marked and sorted rules:\n";
    for (std::vector<Rule *>::iterator i = newRuleVector.begin();
        i != newRuleVector.end(); i++) {
      if ((*i)->targetVector.size())
        for (int j = 0; j < (*i)->targetVector.size(); j++) {
          if (j)
            *outputFile->getOStream() << ' ';
          (*i)->targetVector[j]->outputTo(*outputFile->getOStream());
        }
      else
        *outputFile->getOStream() << " (no target)";
      if ((*i)->sourceVector.size()) {
        *outputFile->getOStream() << " %in";
        for (int j = 0; j < (*i)->sourceVector.size(); j++)
          *outputFile->getOStream() << ' ' << quoteAsNec((*i)->sourceVector[j]->filename);
      }
      else
        *outputFile->getOStream() << " (no sources)";
      if ((*i)->commandVector.size()) {
        *outputFile->getOStream() << " { ";
        for (int j = 0; j < (*i)->commandVector.size(); j++) {
          if (j)
            *outputFile->getOStream() << "; ";
          *outputFile->getOStream() << (*i)->commandVector[j];
        }
        *outputFile->getOStream() << " }\n";
      }
      else
        *outputFile->getOStream() << " (no command)\n";
    }
  }

  int startedSomething = 0;

#define RULESET_RUN_SERIAL    0
#define RULESET_RUN_PARALLEL  1
#define RULESET_RUN_ALGORITHM 1

#if RULESET_RUN_ALGORITHM == RULESET_RUN_PARALLEL
  static int staticMaxPipelines = getenvint("PROCESSORS", get_nprocs());
  if (maxPipelines < 1)
    maxPipelines = std::max(staticMaxPipelines, 1);

  if (otherVerbosity >= 4)
    *outputFile->getOStream() << "[Starting outer loop]" << std::endl;
  for (;;) {
    auto i = newRuleVector.begin();
    if (otherVerbosity >= 4)
      *outputFile->getOStream() << "[Starting inner loop]" << std::endl;
    for (; i != newRuleVector.end(); i++) {
      int result = 0;
      if (otherVerbosity >= 4)
        *outputFile->getOStream() << "Considering " << (*i)->targetVector[0]->filename
          << ' ' << buildStateToMacroName((*i)->getRuleState())
          << ' ' << (*i)->ruleflag << std::endl;
      if ((*i)->getRuleState() == BUILDSTATE_STALE
       || (*i)->getRuleState() == BUILDSTATE_OK
          && ((*i)->ruleflag & RULEFLAG_REBUILD || optionRebuildAll)) {
        if (otherVerbosity >= 1)
          *outputFile->getOStream() << "Starting " << (*i)->targetVector[0]->filename
           << ' ' << buildStateToMacroName((*i)->getRuleState()) << std::endl;
        int resortFlag = 0;
        result = (*i)->startx(resortFlag);
        if (result) {
          if (otherVerbosity >= 4)
            *outputFile->getOStream() << "[Returning " << logValue(result) << "]" << std::endl;

          waitForAllPipelines();

          return result;

        }
        //logMsg << logValue(resortFlag) << std::endl;
        if (resortFlag) {
          logMsg << "Going to resortLabel" << std::endl;
          //lineabort();
          goto resortLabel;
        }
        startedSomething = 1;
        break;
      }
    }
    if (otherVerbosity >= 4)
      *outputFile->getOStream() << "[Finished inner loop]" << std::endl;

    // Current problem: DepTest18 preceding DepTest16 and DepTest22 changes the number of [Waiting
    // for results] lines

    if (i == newRuleVector.end()) {
      if (!pipelines) {
        //logMsg << "Breaking because no pipelines" << std::endl;
        if (otherVerbosity >= 4)
          *outputFile->getOStream() << "[Breaking because no pipelines]" << std::endl;
        break;
      }
      if (otherVerbosity >= 4)
        *outputFile->getOStream() << logValue(pipelines) << std::endl;
      if (otherVerbosity >= 1) {
        *outputFile->getOStream() << "[Waiting for results]" << std::endl;
        //logMsg << "[Waiting for results] " << logValue(pipelines) << ' '
          //<< logValue(maxPipelines) << std::endl;
      }
    }
    else if (pipelines >= maxPipelines) {
      if (otherVerbosity >= 1)
        *outputFile->getOStream() << "[Waiting for a processor]" << std::endl;
    }
    else {
      if (otherVerbosity >= 4)
        *outputFile->getOStream() << "[Continuing]" << std::endl;
      continue;
    }
    int resortFlag = 0; // Added initial value. It's important. Weird it didn't get a compiler error
    //logMsg << logValue(resortFlag) << std::endl;
    int result = waitForRuleExitx(resortFlag, 1);
    //logMsg << logValue(result) << std::endl;
    if (result) {

      // Maybe this should wait for all other pipelines too

      failed = 1;
      if (sharedmem)
        sharedmem->stopnow = 1;
      std::cerr << "Command failed: " << logValue(waitString(result)) << std::endl;
      if (pipelines) {
        waitForAllPipelines();
        /*while (pipelines) {
          std::cerr << "Waiting for " << pipelines << " commands" << std::endl;
          waitForRuleExitx(resortFlag);
        }*/
        std::cerr << "All commands exited" << std::endl;
        std::cerr << "Failed commands were:" << std::endl;
        std::cerr << failedCommands.str();
      }
      if (otherVerbosity >= 3)
        *outputFile->getOStream() << "[returning]" << std::endl;
      return result;
    }
    //logMsg << logValue(resortFlag) << std::endl;
    if (resortFlag) {
      //logMsg << "Going to resortLabel" << std::endl;
      if (otherVerbosity >= 3)
        *outputFile->getOStream() << "[going to resortLabel]" << std::endl;
      goto resortLabel;
    }
    if (otherVerbosity >= 4)
      *outputFile->getOStream() << "[looping around]" << std::endl;
  }

#elif RULESET_RUN_ALGORITHM == RULESET_RUN_SERIAL
  for (auto i = newRuleVector.begin(); i != newRuleVector.end(); i++) {
    int result = 0;
    if (optionRebuildAll || ((*i)->getFileState() == BUILDSTATE_STALE)) {
      if ((*i)->commandVector.size())
        logMsg << "Running " << (*i)->commandVector[0] << std::endl;
      result = (*i)->start();
      startedSomething = 1;
      result = waitForRuleExit();
    }
    else
      if ((*i)->commandVector.size())
        logMsg << "Not running " << (*i)->commandVector[0] << std::endl;
    if (result)
      return result;
  }
  logMsg << "Outputting details to dep.error" << std::endl;
  std::ofstream ofs("dep.error", std::ios::out | std::ios::trunc);
  outputTo(ofs);
  ofs << "Here is the newRuleVector" << std::endl;
  for (auto i = newRuleVector.begin(); i != newRuleVector.end(); i++) {
    if ((*i)->targetVector.size())
      ofs << logValue((*i)->targetVector.front()->filename) << std::endl;
  }
  ofs << "That was the newRuleVector" << std::endl;
#else
  bork bork bork
#endif
  // Maybe check things out here
  int failed1 = failed;
  //logMsg << logValue(failed) << std::endl;
  //logMsg << logValue(failed1) << std::endl;
  auto i = newRuleVector.begin();
  for (; i != newRuleVector.end(); i++) {
    if ((*i)->getRuleState() != BUILDSTATE_OK
     && (*i)->getRuleState() != BUILDSTATE_REBUILT) {
      if (!failed1) {
        *outputFile->getOStream() << "dep: Could not build " << (*i)->targetVector[0]->filename
          << " state is " << buildStateToMacroName((*i)->getRuleState()) << std::endl;
      }
      (*i)->getRuleState(!failed1);
      failed = 1;
    }
  }
  if (!startedSomething && !failed) {
    std::cerr << "dep: all targets are up to date" << std::endl;
  }
  return 0;
}

void RuleSet::waitForAllPipelines() {
  while (pipelines) {
    int resortFlag;
    /*int result =*/ waitForRuleExitx(resortFlag, 0);
    //logMsg << logValue(result) << std::endl;
  }
}

Rule *RuleSet::rulesetAddRule() {
  Rule *rule = new Rule(this);
  rule->ruleflag = 0;
  ruleVector.push_back(rule);
  return rule;
}

BuildFile *RuleSet::getBuildFile(const char *filename) {
  auto itr = fileMap.find(filename);
  if (itr == fileMap.end()) {
    BuildFile *buildfile = new BuildFile(this);
    //logMsg << logValue(filename) << std::endl;
    fileMap[filename] = buildfile;
    buildfile->filename = filename;
    buildfile->fileflag = 0;
    return buildfile;
  }
  return itr->second;
}

void Rule::addTarget(const std::string &filename, int flags, const char *sourceFile, int sourceLine) {
  BuildFile *buildfile = ruleset->getBuildFile(joinFolderAndFile(folder, filename).c_str());
  //logMsg << filename << ' ' << logValue(flags) << std::endl;
  if (flags & FILEFLAG_PSEUDOFILE)
    buildfile->fileflag |= FILEFLAG_PSEUDOFILE;
  if (flags & FILEFLAG_COMPARE)
    buildfile->fileflag |= FILEFLAG_COMPARE;
  if (flags & FILEFLAG_ZEROLENGTH)
    buildfile->fileflag |= FILEFLAG_ZEROLENGTH;
  if (buildfile->writeRule) {
    //logMsg << logValue(ruleset->os) << std::endl;
    if (ruleset->outputFile->getOStream())
      *ruleset->outputFile->getOStream() << sourceFile << ' ' << sourceLine << ": "
        << filename << " already has a rule" << std::endl;
  }
  else
    buildfile->writeRule = this;
  targetVector.push_back(buildfile);
}

int Rule::addSource(const std::string &filename, int flags) {
  BuildFile *buildfile = ruleset->getBuildFile(joinFolderAndFile(folder, filename).c_str());

  if (flags & FILEFLAG_EXIST)
    buildfile->fileflag |= FILEFLAG_EXIST;

  std::vector<Rule *>::iterator i = buildfile->readRuleVector.begin();
  for (; i != buildfile->readRuleVector.end(); i++)
    if (*i == this)
      return 0;
  buildfile->readRuleVector.push_back(this);
  sourceVector.push_back(buildfile);
  return 1;
}

// Tricky thing about includes is that they SOMETIMES depend on the Build Command.
// Because it has the -I options.

int BuildFile::treeAddInclude(const std::string &filename, RuleSet *ruleset, int duplicateCheckIdx) {
  BuildFile *buildfile = ruleset->getBuildFile(filename.c_str());
  //logMsg << this->filename << " includes " << filename << std::endl;
  for (int i = 0; i < includeVector.size(); i++) {
    if (includeVector[i] == buildfile) {
      //logMsg << logValue(i) << ' ' << logValue(duplicateCheckIdx) << std::endl;
      //if (i >= duplicateCheckIdx)
        //std::cerr << this->filename << " already includes " << filename << std::endl;
      return 0;
    }
  }
  buildfile->includedByVector.push_back(this);
  includeVector.push_back(buildfile);
  return 1;
}

void Rule::markRule() {
  if (!(ruleflag & RULEFLAG_MARKED)) {
    // Removing all of these blows the TrinityTests FSR
    //logMsg << logValue(targetVector.size()) << std::endl;
    //logMsg << "Marking " << targetVector[0]->filename << std::endl;
    ruleflag |= RULEFLAG_MARKED;
    for (std::vector<BuildFile *>::iterator j = sourceVector.begin(); j != sourceVector.end(); j++)
      (*j)->markFile();
  }
  //else
    //logMsg << "Not marking " << targetVector[0]->filename << std::endl;
}

void BuildFile::markFile() {
  if (writeRule)
    writeRule->markRule();
  for (std::vector<BuildFile *>::iterator j = includeVector.begin(); j != includeVector.end(); j++)
    (*j)->markFile();
}

/* subincludes.

arg.c: #include "bla.h"
bla.h: #include "coo.h"

arg.c is a source of arg.o
bla.h is a source of arg.o
coo.h is a source of arg.o

But I don't want to list those directly. Instead, I want this:

bla.h is a subinclude of arg.c
coo.h is a subinclude of bla.h

*/

Rule::Rule(RuleSet *ruleset): ruleset(ruleset), ruleflag(0), buildstate(0) {
}

BuildFile::BuildFile(RuleSet *ruleset): writeRule(0), fileflag(0), ruleset(ruleset), includesData(0) {
  fileSize = 0;
  localModTime.tv_sec = 0;
  localModTime.tv_nsec = 0;
  includeModTime.tv_sec = 0;
  includeModTime.tv_nsec = 0;
}

std::string RuleSet::findFile(const std::string &filename, const std::vector<std::string> *folders,
    std::fstream &is, const char *reading, const char *workingDir) const {
  int detailed = 0;
  tryagain:
  //if (detailed)
    //logMsg << "Looking for " << logValue(filename) << std::endl;
  if (folders) {
    for (auto folder: *folders) {
      //std::string longfilename = joinFolderAndFile(folder, filename);
      std::string longfilename = joinFolderAndFile(joinFolderAndFile(workingDir, folder), filename);
      //if (detailed)
        //logMsg << logValue(longfilename) << std::endl;
      is.open(longfilename);
      if (is.is_open())
        return longfilename;
    }
  }
  // Is it a generated file?
  for (auto i = fileMap.begin(); i != fileMap.end(); i++) {
    //if (detailed)
      //logMsg << logValue(i->first) << std::endl;
    if (endsWith(STARTSWITH_FIXED, ("/" + i->first).c_str(), ("/" + filename).c_str())) {
      if (otherVerbosity >= 3)
        *outputFile->getOStream() << "Considering generated file " << logValue(i->first) << std::endl;
      //return i->first;
      for (auto folder: *folders) {
        std::string longfilename = joinFolderAndFile(joinFolderAndFile(workingDir, folder), filename);
        if (longfilename == i->first) {
          if (otherVerbosity >= 3)
            *outputFile->getOStream() << "Returning generated file " << logValue(i->first) << std::endl;
          return i->first;
        }
      }
    }
  }

  // Is it in the same folder as we are reading?
  std::string readingDir = DirName(reading);
  std::string longfilename = joinFolderAndFile(readingDir, filename);
  if (detailed) {
    //logMsg << logValue(reading) << std::endl;
    //logMsg << logValue(DirName(reading)) << std::endl;
    //logMsg << logValue(longfilename) << std::endl;
  }
  is.open(longfilename);
  if (is.is_open())
    return longfilename;

  // Is it in the same folder as we are working?
  longfilename = joinFolderAndFile(workingDir, filename);
  if (detailed) {
    //logMsg << logValue(workingDir) << std::endl;
    //logMsg << logValue(longfilename) << std::endl;
  }
  is.open(longfilename);
  if (is.is_open())
    return longfilename;

  if (!detailed && otherVerbosity >= 0) {
    *outputFile->getOStream() << "Warning: cannot find path of " << filename;
    if (reading && *reading)
      *outputFile->getOStream() << " included by " << reading;
    *outputFile->getOStream() << " (" << logValue(workingDir) << ").\nIf " << filename
      << " is a generated file, it's generating command should go\nabove anything that reads "
      << reading << "\n.Returning the base name " << filename << std::endl;
    // This is irrelevant
    //char cwd[256];
    //getcwd(cwd, 255);
    //*outputFile->getOStream() << "cwd is " << cwd << std::endl;
    if (folders && folders->size()) {
      *outputFile->getOStream() << "Dir list:";
      for (auto folder: *folders)
        if (folder.size())
          *outputFile->getOStream() << ' ' << folder;
      *outputFile->getOStream() << /* "End of dir list" << */ std::endl;
    }
    if (!detailed) {
      detailed = 1;
      goto tryagain;
    }
    //if (verbosity >= 2)
      //*outputFile->getOStream() << "Warning: cannot find path of " << filename
        //<< " while reading " << reading << " in " << workingDir
        //<< ". Returning the base name " << std::endl;
  }
  return filename;
}

void BuildFile::outputTo(std::ostream &os) {
  if (fileflag & FILEFLAG_PSEUDOFILE)
    os << "%pseudo ";
  else if (fileflag & FILEFLAG_COMPARE)
    os << "%cmpout ";
  else if (fileflag & FILEFLAG_ZEROLENGTH)
    os << "%outz ";
  else
    os << "%out ";

  os << quoteAsNec(filename);
}

void RuleSet::outputTo(std::ostream &os) {
  os << '\n';
  for (std::map<std::string, Cmdspec *>::iterator i = syntaxMap.begin();
      i != syntaxMap.end(); i++) {
    os << "syntax " << i->first;
    i->second->outputTo(os);
    os << '\n';
  }
  for (std::vector<Rule *>::iterator i = ruleVector.begin(); i != ruleVector.end(); i++) {
    int started = 0;
    for (auto j: (*i)->targetVector) {
      if (started)
        os << ' ';
      started = 1;
      j->outputTo(os);
    }
    int currentType = -1;
    for (auto j: (*i)->sourceVector) {
      if (j->fileflag & FILEFLAG_EXIST && currentType != FILEFLAG_EXIST) {
        os << " %exist";
        currentType = FILEFLAG_EXIST;
      }
      else if (!(j->fileflag & FILEFLAG_EXIST) && currentType != 0) {
        os << " %in";
        currentType = 0;
      }
      os << ' ' << quoteAsNec(j->filename);
    }
    os << "\n{\n";
    if ((*i)->folder.size())
      os << "cd " << quoteAsNec((*i)->folder) << "\n";
    for (auto j: (*i)->commandVector)
      os << j << '\n';
    os << "}\n\n";
  }
  for (std::map<std::string, BuildFile *>::iterator i = fileMap.begin();
      i != fileMap.end(); i++) {
    os << "fileinfo " << quoteAsNec(i->first);
    switch (i->second->getFileState()) {
    case BUILDSTATE_OK: os << " OK "; break;
    case BUILDSTATE_STALE: os << " STALE "; break;
    case BUILDSTATE_SOURCES_STALE: os << " SOURCES_STALE "; break;
    case BUILDSTATE_REBUILT: os << " REBUILT "; break;
    case BUILDSTATE_REBUILT_SAME: os << " REBUILT_SAME "; break;
    case BUILDSTATE_STAT_DIFF: os << " STAT_DIFF "; break;
    case BUILDSTATE_BUILDING: os << " BUILDING "; break;
    default:
      os << ' ' << buildStateToMacroName(i->second->getFileState()) << ' ' << std::flush;
      logMsg << "No mapping for " << buildStateToMacroName(i->second->getFileState())
        << std::endl;
      lineabort();
    }
    if (i->second->getFileState() == BUILDSTATE_REBUILT
        || i->second->getFileState() == BUILDSTATE_REBUILT_SAME) {
      //logMsg << "Calling calculateModTimes" << std::endl;
      i->second->calculateModTimes();
    }
    if (i->second->fileflag & FILEFLAG_LOCAL_MODTIME_VALID) {
      if (i->second->modTimeResult) {
        os << ErrnoToMacroName(i->second->modTimeResult);
        if (i->second->modTimeResult != ENOENT && i->second->modTimeResult != EINVAL)
          os << ' ';
      }
      if (i->second->modTimeResult != ENOENT && i->second->modTimeResult != EINVAL) {
        os << i->second->localModTime;
        os << ' ' << i->second->fileSize;
      }
    }
    else
      os << "unknown";
    if (i->second->includeVector.size()) {
      os << " includes";
      for (std::vector<BuildFile *>::iterator j = i->second->includeVector.begin();
          j != i->second->includeVector.end(); j++)
        os << ' ' << (*j)->filename;
    }
    os << '\n';
  }
  os << '\n';
}

int operator != (const timespec &a, const timespec &b) {
  return a.tv_sec != b.tv_sec || a.tv_nsec != b.tv_nsec;
}

void RuleSet::compareToCache(const RuleSet &cache) {
  // Go through the files. If there are differences in the file times, mark those files
  int foundARuleChange = 0;
  for (std::map<std::string, BuildFile *>::iterator i = fileMap.begin();
      i != fileMap.end(); i++) {
    auto cacheFile = cache.fileMap.find(i->first);
    if (cacheFile == cache.fileMap.end()) {
      //logMsg << "Couldn't find " << logValue(i->first) << std::endl;
    }
    else {
      //logMsg << "Found " << logValue(i->first) << std::endl;
      //logMsg << "Calling calculateModTimes" << std::endl;
      i->second->calculateModTimes();
      if (i->second->fileflag & FILEFLAG_LOCAL_MODTIME_VALID
          && cacheFile->second->fileflag & FILEFLAG_LOCAL_MODTIME_VALID
          && !i->second->modTimeResult
          && !cacheFile->second->modTimeResult) {
        if (senseModTime && i->second->localModTime != cacheFile->second->localModTime) {
          i->second->fileflag |= FILEFLAG_STAT_CHANGED;
          if (otherVerbosity >= 2)
            *outputFile->getOStream() << i->first << " modtime changed from " << cacheFile->second->localModTime
              << " to " << i->second->localModTime << '\n';
        }
        if (senseSize && i->second->fileSize != cacheFile->second->fileSize) {
          i->second->fileflag |= FILEFLAG_STAT_CHANGED;
          if (otherVerbosity >= 2)
            *outputFile->getOStream() << i->first << " file size changed from " << cacheFile->second->fileSize
              << " to " << i->second->fileSize << '\n';
        }
        if (senseRule) {
          int ruleChanged = 0;
          if (!i->second->writeRule != !cacheFile->second->writeRule) {
            if (!i->second->fileflag & FILEFLAG_EXIST) // ignore rule changes on existence rules
              ruleChanged = 1;
            if (otherVerbosity >= 1 || confirmRuleChanges)
              *outputFile->getOStream() << i->first << " build rule removed\n"
                  << "was: " << cacheFile->second->writeRule->commandVector << '\n';
          }
          else if (i->second->writeRule) {
            if (i->second->writeRule->commandVector.size()
                != cacheFile->second->writeRule->commandVector.size())
              ruleChanged = 1;
            else
              for (int j = 0; j < i->second->writeRule->commandVector.size(); j++)
                if (i->second->writeRule->commandVector[j]
                    != cacheFile->second->writeRule->commandVector[j]) {
                  ruleChanged = 1;
                  break;
                }
          }
          if (ruleChanged) {
            if (otherVerbosity >= 1 || confirmRuleChanges) {
              if (!i->second->writeRule) {
                //*outputFile->getOStream() << i->first << " build rule removed\n"
                    //<< "was: " << cacheFile->second->writeRule->commandVector << '\n';
                foundARuleChange = 1;
              }
              else {
                const char *bnm = " (but not marked)";
                if (i->second->writeRule->ruleflag & RULEFLAG_MARKED) {
                  foundARuleChange = 1;
                  bnm = "";
                }
                if (!strcmp(bnm, "") || otherVerbosity >= 1)
                  if (!cacheFile->second->writeRule)
                    *outputFile->getOStream() << i->first << " build rule added" << bnm << '\n'
                        << "is: " << i->second->writeRule->commandVector << '\n';
                  else
                    *outputFile->getOStream() << i->first << " build rule changed" << bnm << "\n"
                           "from: " << cacheFile->second->writeRule->commandVector << '\n'
                        << "to:   " << i->second->writeRule->commandVector << '\n';
              }
            }
            i->second->fileflag |= FILEFLAG_CMD_CHANGED;
            if (i->second->writeRule)
              i->second->writeRule->invalidateState();
          }
        }
      }
    }
  }
  if (confirmRuleChanges && foundARuleChange) {
    std::cout << "Proceed (Y/n)? " << std::flush;
    std::string reply;
    if (getline(std::cin, reply) && reply.size() && tolower(reply[0]) == 'n')
      exit(0);
  }
}

void RuleSet::combineWithCache(const RuleSet &cache) {
  for (std::map<std::string, BuildFile *>::iterator i = fileMap.begin();
      i != fileMap.end(); i++) {
    auto cacheFile = cache.fileMap.find(i->first);
    if (cacheFile == cache.fileMap.end()) {
      //logMsg << "Couldn't find " << logValue(i->first) << std::endl;
    }
    else {
      //logMsg << "Found " << logValue(i->first) << std::endl;

      // The main question is "Was it executed?" If not, use the cached build rule

      if (i->second->writeRule)
        if (!(i->second->writeRule->buildstate == BUILDSTATE_REBUILT
           || i->second->writeRule->buildstate == BUILDSTATE_REBUILT_SAME)) {
          if (cacheFile->second && cacheFile->second->writeRule)
            i->second->writeRule->commandVector = cacheFile->second->writeRule->commandVector;
        }
    }
  }
}

void BuildFile::treeAddIncludes(const std::string &filename,
    const std::vector<std::string> *folders, std::fstream &is, const char *workingDir) {
  std::ostream &os(*ruleset->outputFile->getOStream());
  //logMsg << "Checkpoint" << std::endl;

  if (this->folders.size()) {
    if (this->folders != *folders)
      //logMsg << "Ooh that's different " << logValue(filename)
        //<< ' ' << logValue(*folders) << std::endl;
      fileflag &= ~FILEFLAG_FOUND_INCLUDES;
      this->folders = *folders;
  }
  else
    this->folders = *folders;

  if (fileflag & FILEFLAG_FOUND_INCLUDES) {
    //logMsg << "Have already looked for includes in " << filename << ". Returning" << std::endl;
    return;
  }
  //logMsg << "Checkpoint" << std::endl;
  fileflag |= FILEFLAG_FOUND_INCLUDES;
  if (ruleset->otherVerbosity >= 4) {
    //logMsg << logValue(&os) << std::endl;
    os << "Looking for includes in " << filename << std::endl;
  }
  //logMsg << "Checkpoint" << std::endl;
  if (!is.is_open()) {
    //logMsg << "Can't read includes from " << filename << std::endl;
    includesData = new IncludesData;
    includesData->filename = filename;
    includesData->workingDir = workingDir;
    return;
  }
  std::string line;
  int n = 0;

  static Pattern includePattern("#include  *\".*\r?", std::regex_constants::ECMAScript); // Need the \r for DOS CRLF files
  static Pattern definePattern("#define.*\r?", std::regex_constants::ECMAScript); // Need the \r for DOS CRLF files

  //logMsg << logValue(ruleset->maxNonIncludes) << std::endl;
  //logMsg << "Checkpoint" << std::endl;
  int duplicateCheckIdx = 0;
  while (getline(is, line) && n < ruleset->maxNonIncludes) {
    if (matches(line, includePattern)) {
      if (ruleset->otherVerbosity >= 5)
        os << "This line matches: " << line << std::endl;
      int quote1 = line.find('"') + 1;
      int quote2 = line.find('"', quote1);
      std::string includeFilename = line.substr(quote1, quote2 - quote1);
      std::fstream is;

      std::string longfilename = ruleset->findFile(includeFilename,
        folders, is, filename.c_str(), workingDir);
      if (ruleset->otherVerbosity >= 4) {
        os << filename << " includes " << longfilename << std::endl;
        //logMsg << logValue(*folders) << std::endl;
        //logMsg << logValue(workingDir) << std::endl;
      }
      // Need to know exactly where this is - what folder
      // Is it a currently-existing file?
      // Or will it be generated? (We don't have to read it if that's the case)

      if (treeAddInclude(longfilename, ruleset, duplicateCheckIdx)) {
        includeVector.back()->treeAddIncludes(longfilename, folders, is, workingDir);
      }
      n = 0;
    }
    else if (matches(line, definePattern)) {
      duplicateCheckIdx = includeVector.size();
      //logMsg << "Checkpoint " << logValue(duplicateCheckIdx) << ' ' << logValue(line) << std::endl;
    }
    else {
      if (ruleset->otherVerbosity >= 5)
        os << "This line does not match: " << line << std::endl;
      n++;
    }
  }
  //logMsg << "Checkpoint" << std::endl;
}

// This version adds sources to a Rule.

void Rule::ruleAddIncludes(const std::string &filename, const std::vector<std::string> *folders,
    std::fstream &is, const std::string &dir) {
  if (ruleset->otherVerbosity >= 3) {
    *ruleset->outputFile->getOStream() << "Looking for includes in " << filename << std::endl;
  }
  if (!is.is_open()) {
    if (ruleset->otherVerbosity >= 2)
      *ruleset->outputFile->getOStream() << "Can't read includes from " << filename << std::endl;
    return;
  }
  std::string line;
  int n = 0;
  static Pattern includePattern("#include  *\".*", std::regex_constants::ECMAScript);
  //logMsg << logValue(ruleset->maxNonIncludes) << std::endl;
  while (getline(is, line) && n < ruleset->maxNonIncludes) {
    if (matches(line, includePattern)) {
      int quote1 = line.find('"') + 1;
      int quote2 = line.find('"', quote1);
      std::string includeFilename = line.substr(quote1, quote2 - quote1);
      std::fstream is;
      if (ruleset->otherVerbosity >= 2)
        *ruleset->outputFile->getOStream() << BaseName(__FILE__) << " X522 Calling findFile on " << includeFilename
          << std::endl;
      std::string longfilename = ruleset->findFile(includeFilename,
        folders, is, filename.c_str(), folder.c_str());
      //logMsg << filename << " includes " << longfilename << std::endl;
      // Need to know exactly where this is - what folder
      // Is it a currently-existing file?
      // Or will it be generated? (We don't have to read it if that's the case)

      if (addSource(longfilename, 0)) {
        if (ruleset->otherVerbosity >= 2)
          *ruleset->outputFile->getOStream() << BaseName(__FILE__) << " X533 Calling ruleAddIncludes on "
            << logValue(longfilename) << std::endl;
        ruleAddIncludes(longfilename, folders, is, dir);
      }
      n = 0;
    }
    else
      n++;
  }
}

BuildFile::~BuildFile() {
  delete includesData;
}

